'use strict'
const util = require('util');
var DBusProxy = require("./DBusProxy.js");
var log = require("../../utils/log");
var constant = require("../../utils/constant");
const TAG = "[CMPProxy]";
var CMPProxy = function () {
  this.proxies = {};

  this.makeProxy = function (id) {
    var proxy = this.proxies[id];
    if (!proxy) {
      var wellknow = util.format("iot.device.id%s", id);
      var objpath = util.format("/iot/device/id%s", id);
      var iface = util.format("iot.device.id%s", id);
      proxy = new DBusProxy(wellknow, objpath, iface);
      this.proxies[id] = proxy;
    }
    return proxy;
  };

  /**
   * "{\"code\": 0, \"params\": {\"code\": 96, \"message\": \"YellowLight: type is invalid\", \"data\": {}}, \"message\": \"successfully\"}"
   * @param str
   * @param prompt
   * @return {*}
   */
  this.parseReturnValue = function (str, prompt) {
    try {
      let result = JSON.parse(str);
      return result;
    } catch (e) {
      console.log(util.format("[CMPProxy.%s] parse return value failed: %s"), prompt, e.message);
    }

    return undefined;
  }
};

CMPProxy.prototype.getProperty = function (id, property) {
  var args = {
    "params": [property]
  };
  log.i(TAG, 'getProperty:', id, property);
  var value = this.makeProxy(id).callService('callServices', 'get', JSON.stringify(args));
  log.i(TAG, util.format("CMPProxy.getProperty('%s.%s') = %s", id, property, value));
  var result = this.parseReturnValue(value, 'getProperty');
  if (result !== undefined && result.code === 0) {
    console.log("read: " + JSON.stringify(result.params));
    return result.params[property];
  } else {
    return undefined;
  }
};

CMPProxy.prototype.getPropertyAsync = function (id, property, callback) {
  var args = {
    "params": [property]
  };
  log.i(TAG, 'getPropertyAsync:', id, property);
  var self = this;
  this.makeProxy(id).callServiceAsync('callServices', 'get', JSON.stringify(args), function (err, value) {
    log.i(TAG, 'getPropertyAsync return value: id:', id, 'err', err, 'value:', value);
    if (err) {
      callback(err, undefined);
    } else {
      var result = self.parseReturnValue(value, 'getProperty');
      if (result !== undefined) {
        if (result.code === 0) {
          callback(undefined, result.params[property]);
        } else {
          callback(new Error(result.message), undefined);
        }
      } else {
        callback(new Error("getProperty failed for internal error."), undefined);
      }
    }
  });
};

CMPProxy.prototype.getBatchPropertiesAsync = function (id, properties, callback) {
  var args = {
    "params": properties
  };
  log.i(TAG, 'getBatchPropertiesAsync:', id, JSON.stringify(properties));
  var self = this;
  this.makeProxy(id).callServiceAsync('callServices', 'get', JSON.stringify(args), function (err, value) {
    log.i(TAG, 'getBatchPropertiesAsync return value: id:', id, 'err', err, 'value:', value);
    if (err) {
      callback(err, undefined);
    } else {
      var result = self.parseReturnValue(value, 'getBatchPropertiesAsync');
      if (result !== undefined) {
        if (result.code === 0) {
          callback(undefined, result.params);
        } else {
          callback(new Error(result.message), undefined);
        }
      } else {
        callback(new Error("getBatchPropertiesAsync failed for internal error."), undefined);
      }
    }
  });
};

CMPProxy.prototype.setProperty = function (id, property, value) {
  var args = {
    "params": {}
  };
  log.i(TAG, 'setProperty:', id, property, value);
  args.params[property] = value;
  var retmsg = this.makeProxy(id).callService('callServices', 'set', JSON.stringify(args));
  var result = this.parseReturnValue(retmsg, 'setProperty');
  log.i(TAG, "setProperty : " + retmsg);
  return result == null ? false : result.code === 0;
};

/**
 * /**
 * "{\"code\": 0, \"params\": {\"code\": 96, \"message\": \"YellowLight: type is invalid\", \"data\": {}}, \"message\": \"successfully\"}"
 * @param str
 * @param prompt
 * @return {*}
 * @param id
 * @param property
 * @param val
 * @param callback
 */
CMPProxy.prototype.setPropertyAsync = function (id, property, val, callback) {
  var args = {
    "params": {}
  };
  log.i(TAG, 'setPropertyAsync:', id, property, val);
  args.params[property] = val;
  var self = this;
  this.makeProxy(id).callServiceAsync('callServices', 'set', JSON.stringify(args), function (err, value) {
    log.i(TAG, 'setPropertyAsync return value: id:', id, 'err', err, 'value:', value);
    if (err) {
      callback(err, undefined);
    } else {
      var obj = self.parseReturnValue(value, 'setPropertyAsync');
      if (obj && obj.code === 0) {
        callback(undefined, true);
      } else {
        callback(undefined, false);
      }
    }
  });
};

/**
 * 一次设置设备的一批属性
 * @param {[type]}   id         [设备ID]
 * @param {[type]}   properties [一个Object对象，包含要设置的参数，格式如下：
 *                                  ｛
 *                                      key1: value1,
 *                                      key2: value2,
 *                                      ...
 *                                  ｝
 *                              ]
 * @param {Function} callback   [description]
 */
CMPProxy.prototype.setBatchPropertiesAsync = function (id, properties, callback) {
  var args = {
    "params": properties
  };
  log.i(TAG, 'setBatchPropertiesAsync:', id, JSON.stringify(args));
  var self = this;
  this.makeProxy(id).callServiceAsync('callServices', 'set', JSON.stringify(args), function (err, value) {
    log.i(TAG, 'setBatchPropertiesAsync return value: id:', id, 'err', err, 'value:', value);
    if (err) {
      callback(err, undefined);
    } else {
      var obj = self.parseReturnValue(value, 'setBatchPropertiesAsync');
      if (obj && obj.code === 0) {
        callback(undefined, true);
      } else {
        callback(undefined, false);
      }
    }
  });
};

/**
 * 按照与CMP的约定，DBUS上为了保持与云端上同步，返回的数据格式如下：
 * {"message":"successfully",
 * 　"code":0,　// 表示DBUS的操作成功与否
 * 　"params":{
 * 　    "message":"success",
 * 　    "code":0,　// 业务操作是否完成
 * 　    "data":{"state":"-1"}}}
 * @param  {[type]} id     [description]
 * @param  {[type]} method [description]
 * @param  {[type]} args   [description]
 * @return {[type]}        [description]
 */
CMPProxy.prototype.callService = function (id, method, args) {
  var params = {params: args};
  log.i(TAG, 'callService:', id, method, JSON.stringify(params));
  var value = this.makeProxy(id).callService('callServices', method, JSON.stringify(params));
  var obj = this.parseReturnValue(value, 'callService') || undefined;
  log.i(TAG, "CMPProxy callService return value: ", value);
  if (obj !== undefined && obj.code === 0) {
    var payload = obj.params;
    if (payload.code === 0) {
      return payload.data;
    }
  }
};

CMPProxy.prototype.callServiceAsync = function (id, method, args, callback) {
  var self = this;
  var params = {params: args};
  log.i(TAG, 'callServicesAsync:', id, method, JSON.stringify(params));
  this.makeProxy(id).callServiceAsync('callServices', method, JSON.stringify(params), function (err, value) {
    var obj = self.parseReturnValue(value, 'callServiceAsync');
    log.i(TAG, 'callServiceAsync return value: id:', id, 'method:', method, 'value:', JSON.stringify(value));
    if (err) {
      callback(err, undefined);
    } else {
      if (obj !== undefined) {
        if (obj.code === 0) {
          var payload = obj.params;
          // 有些方法没有返回值，params是一个空字符, 如果data是空，就使用code字段
          //
          /*
          {
              "code": 0,
              "message":  "ok",
              "params":   {
                  "code": 0,
                  "msg":  "ok",
                  "data":   ""
              }
          }
           */
          if (typeof payload !== 'object') {
            callback(undefined, true);
          } else {
            if (payload.code === 0) {
              var pdata = payload.data;
              if (pdata == null || (typeof pdata === 'string' && pdata.length === 0)) {
                callback(undefined, 0);
              } else {
                callback(undefined, payload.data);
              }
            } else {
              callback(new Error(payload.message), undefined);
            }
          }
        } else {
          callback(new Error(obj.message), undefined);
        }
      } else {
        callback(new Error("get return value is undefined."), undefined);
      }
    }
  });
};

module.exports = new CMPProxy();
