/*
 * Copyright (c) 2019 Alibaba Group Holding Ltd. All rights reversed.
 */


// Log to stdout for debugging when started standalone.
console.log('Start IFTTT...');

const Log = require('./utils/log');

// Setup global exception handler.
process.on('uncaughtException', function (err) {
  Log.e('Uncaught exception: ', err.stack);
  // Exit after 3s so that logs can be saved.
  setTimeout(function () {
    process.exit(-1);
  }, 3000);
});

// #################### IFTTT BEGIN ######################
const fs = require('fs');
const path = require('path');
const consts = require("./utils/constant");
const DBusProxy = require('./dbus/dbus-proxy/DBusProxy');
const Scene = require('./scene.js');
const condition = require("./condition");

const CONF_VERSION = 2;
const CONF_PATH = path.join(consts.FOTA_IFTTT, 'configuration.json');

const ACTION_ADD_SCENE = 10201; // 增加场景
const ACTION_DELETE_SCENE = 10204; // 删除场景
const ACTION_DELETE_JOB = 10211; //冗余场景删除

var IFTTT = function (scheduler, eventBus, stateMachine) {
  this.scenes = [];
  this.jobInfos = []; // Deployment infos
  this.scheduler = scheduler;
  this.eventBus = eventBus;
  this.stateMachine = stateMachine;
};

/**
 * Load scenes from config-manager, and load running persistent configs from local
 * configuration file(configuration.json) if it exists.
 */
IFTTT.prototype.loadScenes = function (callback) {
  // Add default callback.
  var cbk = callback || function () {};
  var self = this;
  var configuration;
  try {
    Log.i('Load configuration from local file at %s.', CONF_PATH);
    var data = fs.readFileSync(CONF_PATH);
    configuration = JSON.parse(data);
    Log.i('Configuration loaded from local file: ', JSON.stringify(configuration));
  } catch (err) {
    if (err.code === 'ENOENT') {
      // No running persistent configs.
      Log.i('No such local configuration file: ' + CONF_PATH);
    } else {
      // Read file fails, discard the configs.
      Log.i('Fail to read local file: %s: %s. Discard the configuration in local ' +
        'configuration file.', CONF_PATH, err.message);
    }
  }

  // Add version to configuration if it's null or its version it null, or upgrade the
  // version to 2 if the version is less than 2.
  if (!configuration || !configuration.version || configuration.version < 2) {
    configuration = {
      version: CONF_VERSION
    };
  }

  try {
    Log.i('Load scenes from config-manager...');
    var arrayStr = DBusProxy.proxy.getConfig(consts.MODULE_NAME);
    if (!arrayStr) {
      Log.i('No scenes are loaded from config-manager, skip adding...');
      return;
    }
    var jobInfos = JSON.parse(arrayStr);
    if (!Array.isArray(jobInfos)) {
      throw new Error('Scenes loaded from config-manager are not in a array.');
    }

    this.jobInfos = jobInfos;
    Log.i('Add scenes: ', arrayStr);
    var count = 0;
    jobInfos.forEach(function (info) {
      self.addScene(info, function (err, scene) {
        count++;
        if (!err) {
          // Merge running persistent configs.
          var persistent = configuration[scene.sceneid];
          if (persistent) {
            scene.enabled = persistent.enabled;
          }
        }
        if (count === jobInfos.length) {
          Log.i('Add scenes complete.');
          cbk(undefined);
          self.saveScenes();
        }
      });
    });
  } catch (e) {
    Log.w('Load and parse scenes from config-manager failed: ', e);
    cbk(e);
  }
};

/**
 * Save scenes and running persistent configs to local configuration
 * file(configuration.json).
 */
IFTTT.prototype.saveScenes = function () {
  try {
    var toSaves = {
      version: CONF_VERSION
    };
    this.scenes.forEach(function (scene) {
      // Duplicate sceneObj.
      var replica = JSON.parse(JSON.stringify(scene.sceneObj));
      replica.enabled = scene.enabled;
      toSaves[scene.sceneid] = replica;
    });
    Log.i('Save configuration to local file %s', CONF_PATH);
    fs.writeFileSync(CONF_PATH, JSON.stringify(toSaves));
  } catch (err) {
    Log.w('Save configuration failed: ' + err.message);
  }
};

IFTTT.prototype.saveJobInfos = function () {
  Log.i('Save job infos: %s', this.jobInfos.map(function (jobInfo) {
    return jobInfo.resourceId;
  }));
  return DBusProxy.proxy.setConfig(consts.MODULE_NAME, JSON.stringify(this.jobInfos));
};

IFTTT.prototype.addJobInfo = function (info) {
  if (!info || !info.resourceId) {
    return false;
  }
  Log.i('Add job info for %s', info.resourceId);

  for (var i = 0; i < this.jobInfos.length; i++) {
    if (this.jobInfos[i].resourceId === info.resourceId) {
      this.jobInfos.splice(i, 1);
      break;
    }
  }
  this.jobInfos.push(info);
  return true;
};

IFTTT.prototype.deleteJobInfo = function (info) {
  if (!info || !info.resourceId) {
    return;
  }
  Log.i('Delete job info for %s', info.resourceId);

  for (var i = 0; i < this.jobInfos.length; i++) {
    if (this.jobInfos[i].resourceId === info.resourceId) {
      this.jobInfos.splice(i, 1);
      return info;
    }
  }
  return;
};

IFTTT.prototype.deleteAllJobInfos = function () {
  Log.i('Deleting all job infos...');
  this.jobInfos = [];
  return true;
};

IFTTT.prototype.enableScene = function (sceneId, callback) {
  var scene = this.scenes.find(function (scene) {
    return sceneId === scene.id;
  });
  if (!scene) {
    return callback(new Error('Scene does not exist, id = ' + sceneId));
  }
  scene.enabled = true;
  this.saveScenes();
  callback(undefined);
};

IFTTT.prototype.disableScene = function (sceneId, callback) {
  var scene = this.scenes.find(function (scene) {
    return sceneId === scene.id;
  });
  if (!scene) {
    return callback(new Error('Scene does not exist, id = ' + sceneId));
  }
  scene.enabled = false;
  this.saveScenes();
  callback(undefined);
};

IFTTT.prototype.addScene = function (info, callback) {
  Log.i('Adding scene: ', JSON.stringify(info));

  var self = this;
  var sdlPath = info.content && info.content.file;

  fs.readFile(sdlPath, { encoding: 'utf-8' }, function (err, sdlStr) {
    if (err) {
      Log.i('Fail to read %s: %s.', sdlPath, err.stack);
      return callback(err);
    }

    var obj;
    try {
      obj = JSON.parse(sdlStr);
    } catch (ex) {
      Log.i('Parse scene failed: %s: %s.', sdlStr, ex.stack);
      return callback(ex);
    }

    var scene = new Scene(self.scheduler, self.eventBus, self.stateMachine);
    scene.jobInfo = info;
    if (!scene.parseScene(obj)) {
      var message = 'Illegal scene definition: ' + sdlStr;
      Log.i(message);
      return callback(new Error(message));
    }

    self.scenes.push(scene);
    Log.i('Scene added: ' + scene.sceneid);
    callback(undefined, scene);
  });
};

IFTTT.prototype.deleteScene = function (info) {
  Log.i('Deleting scene: ', JSON.stringify(info));
  var isSuccess = false;
  var found = false;
  var response = {};
  response.resourceId = info.resourceId;
  response.jobStatus = 0;
  response.resourceName = info.resourceName;
  var tag = '[IFTTT_Delete_Result][Succeeded]';
  response.msg = {};
  response.msg.code = 0;
  for (var i = 0; i < this.scenes.length; ++i) {
    if (this.scenes[i].jobInfo.resourceId === info.resourceId) {
      isSuccess = this.scenes[i].stop();
      if (isSuccess) {
        this.scenes.splice(i, 1);
        found = true;
        response.msg.detail = 'delete success for resourceId:' + info.resourceId;
      }
      break;
    }
  }
  if (!found) {
    response.msg.detail = 'delete success for not existed resourceId:' + info.resourceId;
  }
  Log.m(tag, '[' + response.resourceName + ']', 'deleteScene result = ', JSON.stringify(response));
};

IFTTT.prototype.deleteAllScenes = function () {
  Log.i('Deleting all scenes...');

  for (var i = this.scenes.length - 1; i >= 0; i--) {
    var scene = this.scenes[i];
    var success = scene.stop();
    if (success) {
      this.scenes.splice(i, 1);
    }
  }
};

IFTTT.prototype.run = function () {
  Log.i('Start IFTTT...');

  this.loadScenes();

  this.eventBus.subscribe('propertiesChanged', this.onPropertiesChanged.bind(this));
  this.eventBus.subscribe('thingEventHappened', this.onThingEventHappened.bind(this));
  this.eventBus.subscribe('timerEvent', this.onTimerEvent.bind(this));
  this.eventBus.subscribe('invokeScene', this.onInvokeScene.bind(this));
  this.eventBus.subscribe('updateScenes', this.onUpdateScenes.bind(this));
  this.eventBus.subscribe('resetScenes', this.onResetScenes.bind(this));

  // Setup WebSocket server for requests from Action processes.
  var webSocketServer = require('./dbus/WSServer');
  webSocketServer.setIFTTT(this);

  // Setup DBus services
  require('./dbus');

  Log.i('IFTTT is running...');
};

IFTTT.prototype.onPropertiesChanged = function (event) {
  this.scenes.forEach(function (scene) {
    scene.propertiesChanged(event);
  });
};

IFTTT.prototype.onThingEventHappened = function (event) {
  var info = event;
  this.scenes.forEach(function (scene) {
    scene.eventHappened(info);
  });
};

IFTTT.prototype.onTimerEvent = function (event) {
  var info = event;
  Log.i('onTimerEvent: ' + JSON.stringify(info));
  this.scenes.forEach(function (scene) {
    scene.timerTriggered(info);
  });
};

IFTTT.prototype.onUpdateScenes = function (info) {
  var self = this;
  try {
    var resources = JSON.parse(info.resources);
    for (var i = 0; i < resources.length; i++) {
      var resource = resources[i];
      if (ACTION_ADD_SCENE === info.ackCode) {
        Log.i('onUpdateScenes action=ACTION_ADD_SCENE');
        // Add the job info and save.
        self.addJobInfo(resource);
        self.saveJobInfos();

        // Add the scene corresponding to the job info and save.
        self.deleteScene(resource);
        self.addScene(resource, function (err) {
          if (err) {
             Log.i('Add scene failed: %s: %s', err.message, JSON.stringify(resource));
             return;
          }
          self.saveScenes();
        });
      } else if (ACTION_DELETE_SCENE === info.ackCode || ACTION_DELETE_JOB === info.ackCode) {
        Log.i('onUpdateScenes action=ACTION_DELETE_SCENE');
        // Delete the job info and save.
        self.deleteJobInfo(resource);
        self.saveJobInfos();

        // Delete the scene corresponding to the job info and save.
        self.deleteScene(resource);
        self.saveScenes();
      } else {
        Log.i('onUpdateScenes unknown action=%s', info.ackCode);
      }
    }
  } catch (e) {
    Log.w('Not a JSON: ', e.stack);
  }
};

IFTTT.prototype.onResetScenes = function () {
  this.deleteAllJobInfos();
  this.saveJobInfos();

  this.deleteAllScenes();
  this.saveScenes();

  Log.i('Reset scenes complete.');
};

IFTTT.prototype.onInvokeScene = function (event) {
  Log.i('onInvokeScene event=%s', JSON.stringify(event));
  var iface = event['interface'];
  if(iface === 'actionOnce') {
    var payload = JSON.parse(event['payload']);
    var resourceId = payload['sceneId'];
    this.scenes.forEach(function (scene) {
      if(scene.jobInfo.resourceId === resourceId) {
        condition.isConditionsMeeted((scene.sceneObj)['condition'], function (err, satisfy) {
          Log.i('Meet conditions: sid=%s, err=%s, meet=%s', resourceId, err, satisfy);
          if(!err && satisfy) {
            scene.executeJob();
          }
        });
      }
    });
  }
};
// #################### IFTTT END ######################

const broker = require('./broker/index');
const scheduler = require("./scheduler.js");
new IFTTT(scheduler, broker).run();