/**
 * "action":[
        {
            "node":"action/device/setStatus",
            "params":{
                "productKey":"",
                "deviceName":"",
                "propertyItems":{
                    "attr":"value",
                    "attr2":"value2"
                }
            }
        }
    ]
 */
const Tag = require('../utils/constant');
const Log = require('../utils/log');
const fs = require('fs');
const path = require('path');
const LOGTAG = "[Action]";
const util = require('util');
var condition = require("./condition");
const { PROJECT_PATH } = require('../const');

var Action = function(actions, sid, proxy, traceId) {
    this.sceneId = sid;
    this.proxy = proxy;
    this.actions = actions;
    this.traceId = traceId;
};

/** DETAIL 日志 ['uri', 'bizCode', 'traceId', 'sceneId', 'code', 'msg', 'timestamp', ...,
 * 'productKey', 'deviceName', 'propertyItems', 'serviceName', 'serviceArgs']
 * [2018-05-02 17:46:00]; uri:action/device/setProperty; bizCode:DETAIL; traceId:0a6550ce15252543600081930d1767; sceneId:a111e96fe83b44348e8a7ed9864a0c48; code:200; msg:success; timestamp:1525254360452; productKey:a1NfftnCLGg; deviceName:wudSNSEccE0ppPhKi88X; propertyItems:{"int_rw":5}
 */
Action.prototype.generateLog = function (code, msg, uri, productKey, deviceName, propertyItems, serviceName, serviceArgs) {
    var log = util.format("[%s]", (new Date).toLocaleString());
    log += ";uri:" + uri;
    log += ";bizCode:" + "DETAIL";
    log += ";traceId:" + this.traceId;
    log += ";sceneId:" + this.sceneId;
    log += ";code:" + code;
    log += ";msg:" + msg.toString();
    log += ";timestamp:" + Math.round(new Date().getTime());// 毫秒
    if(productKey) {
        log += ";productKey:" + productKey;
    }
    if(deviceName) {
        log += ";deviceName:" + deviceName;
    }
    if(propertyItems) {
        log += ";propertyItems:" + JSON.stringify(propertyItems);
    }
    if(serviceName) {
        log += ";serviceName:" + serviceName;
    }
    if(serviceArgs) {
        log += ";serviceArgs:" + JSON.stringify(serviceArgs);
    }
    log += "\n";
    return log;
};

Action.prototype.execute = function (cb) {
    Log.i(LOGTAG, "execute begin sid = " + this.sceneId);
    var self = this;
    var actionsLog = [];
    if(!this.actions || this.actions.length === 0) return cb(new Error('actions empty'));
    var hasReturned = false;
    var cnt = 0;
    var returnErr = false;
    var maxDelay = 0;
    var processTimeout = undefined;
    function callback(err, actionLog) {
        if (processTimeout) {
            clearTimeout(processTimeout);
            processTimeout = undefined;
        }
        return cb(err, actionLog);
    }
    Log.i(LOGTAG, "execute actions size = ", this.actions.length);
    this.actions.forEach(function (action) {
        if (!action || !(action)[Tag.PARAMS] || !action[Tag.URI]) {
            var msg = "action sid " + self.sceneId + " params error, action = " + JSON.stringify(action);
            Log.w(LOGTAG, msg);
            actionsLog.push(self.generateLog(Tag.RUN_FAIL, msg, null, null, null, null, null, null));
            cnt++;
            returnErr = true;
            return;
        }
        var params = action[Tag.PARAMS];
        var actionUri = action[Tag.URI];
        var delay = params['delayedExecutionSeconds'] || 0;
        if (maxDelay < delay) {
            maxDelay = delay;
        }
        Log.i(LOGTAG, "execute actionUri = " + actionUri);
        if (actionUri === Tag.ActionSetDevice || actionUri === Tag.ActionSetProperty) {
            var pk = params[Tag.PRODUCTKEY];
            var dn = params[Tag.DEVICENAME];
            var id = pk + "_" + dn;
            var attr = params[Tag.PROPERTYITEMS];
            if (attr) {
                function setProperty() {
                    Log.i(LOGTAG, "setProperty", id, JSON.stringify(attr));
                    self.proxy.setProperty(id, attr, function(err, value) {
                        Log.i(LOGTAG, 'execute sid =', self.sceneId, actionUri, 'attr', JSON.stringify(attr), 'result: err =', err, 'value =', JSON.stringify(value));
                        var code = Tag.RUN_SUCCESS;
                        var msg = "success";
                        if(err) {
                            returnErr = err;
                            code = Tag.RUN_FAIL;
                            msg = err.toString();
                        }
                        actionsLog.push(self.generateLog(code, msg, action[Tag.URI], pk, dn, attr, null, null));
                        cnt++;
                        if(!hasReturned && (cnt === self.actions.length)) {
                            hasReturned = true;
                            return callback(returnErr, actionsLog);
                        }
                    });
                }
                Log.i(LOGTAG, "Set property in " + delay + " seconds.");
                setTimeout(setProperty, delay * 1000);
            } else {
                cnt++;
                var mess = 'propertyItems for action is not defined.';
                Log.w(LOGTAG, mess);
                returnErr = true;
                actionsLog.push(self.generateLog(Tag.RUN_FAIL, mess, action[Tag.URI]), pk, dn, attr, null, null);
            }
        } else if (action[Tag.URI] === Tag.ActionInvokeService) {
            var pk = params[Tag.PRODUCTKEY];
            var dn = params[Tag.DEVICENAME];
            var id = pk + "_" + dn;
            var srvName = params[Tag.ServiceName];
            var srvArgs = params[Tag.ServiceArgs];
            if (srvName && srvArgs) {
                function invokeService() {
                    Log.i(LOGTAG, "callService", id, srvName, JSON.stringify(srvArgs));
                    self.proxy.callService(id, srvName, srvArgs, function(err, value) {
                        Log.i(LOGTAG, 'execute sid = ', self.sceneId, Tag.ActionInvokeService, 'result: err = ', err, 'value =', JSON.stringify(value));
                        var code = Tag.RUN_SUCCESS;
                        var msg = "success";
                        if(err) {
                            returnErr = err;
                            code = Tag.RUN_FAIL;
                            msg = err.toString();
                        }
                        actionsLog.push(self.generateLog(code, msg, action[Tag.URI], pk, dn, null, srvName, srvArgs));
                        cnt++;
                        if(!hasReturned && (cnt === self.actions.length)) {
                            hasReturned = true;
                            return callback(returnErr, actionsLog);
                        }
                    });
                }
                Log.i(LOGTAG, "Invoke service in " + delay + " seconds.");
                setTimeout(invokeService, delay * 1000)
            } else {
                cnt++;
                var mess = 'action execute　sid = ' + self.sceneId + action[Tag.URI] + ' failed for: ' + Tag.ServiceName + '=' + srvName +  Tag.ServiceArgs + " = " + srvArgs;
                Log.w(LOGTAG, mess);
                returnErr = true;
                actionsLog.push(self.generateLog(Tag.RUN_FAIL, mess, action[Tag.URI], pk, dn, null, srvName, srvArgs));
            }
        } else if (action[Tag.URI] === Tag.ActionScence) {
            var sid = params[Tag.SCENEID];
            Log.i(LOGTAG, Tag.ActionScence, 'findTargetScene begin sid =', sid);
            findTargetScene(sid, function (err, sceneObject) {
                var mess = 'findTargetScene ' +  sid +  ' err =' + err + " sceneObject = " + JSON.stringify(sceneObject);
                Log.i(LOGTAG, mess);
                if (err || !sceneObject) {
                    cnt++;
                    Log.w(LOGTAG, mess);
                    returnErr = err;
                    actionsLog.push(self.generateLog(Tag.RUN_FAIL, 'sid ' + sid + " is not found, err = " + err.toString(), action[Tag.URI], null, null, null, null, null));
                    if(!hasReturned && (cnt === self.actions.length)) {
                        hasReturned = true;
                        return callback(returnErr, actionsLog);
                    }
                } else {
                    var conditionObject = sceneObject[Tag.CONDITION] || [];
                    Log.i(LOGTAG, "isConditionsMeeted condition = " + JSON.stringify(conditionObject));
                    condition.isConditionsMeeted(conditionObject, function (err, satisfy) {
                        Log.i(LOGTAG, "isConditionsMeeted err = " + err + " satisfy =" + satisfy);
                        if (!err && satisfy) {
                            Log.i(LOGTAG, "Subscene condition satisfy begin to execute");
                            var actionInner = new Action(sceneObject[Tag.ACTION], self.sceneId, self.proxy, self.traceId);
                            actionInner.execute(function (err, status) {
                                Log.i(LOGTAG, "actionInner execute callback err = " + err + " status =" + status);
                                cnt++;
                                var code = Tag.RUN_SUCCESS;
                                var mess = "success";
                                if(err) {
                                    returnErr = err;
                                    code = Tag.RUN_FAIL;
                                    mess = err.toString();
                                    Log.w(LOGTAG, mess);
                                }
                                //actionsLog.push(self.generateLog(code, mess, action[Tag.URI], null, null, null, null, null)); quit subscene report itself status
                                actionsLog.push(status);
                                if(!hasReturned && (cnt === self.actions.length)) {
                                    hasReturned = true;
                                    return callback(returnErr, actionsLog);
                                }
                            });
                        } else if(err) {
                            cnt++;
                            var mess = 'isConditionsMeeted err = ' + err.toString();
                            Log.w(LOGTAG, mess);
                            returnErr = err;
                            actionsLog.push(self.generateLog(Tag.RUN_FAIL, mess, action[Tag.URI]));
                            if(!hasReturned && (cnt === self.actions.length)) {
                                hasReturned = true;
                                return callback(returnErr, actionsLog);
                            }
                        } else if(!satisfy) {
                            Log.i(LOGTAG, "sid", sid, "isConditionsMeeted  satisfy =" + satisfy, "will not execute");
                            cnt++;
                            if(!hasReturned && (cnt === self.actions.length)) {
                                hasReturned = true;
                                return callback(returnErr, actionsLog);
                            }
                        }
                    });
                }
            });
        } else if (action[Tag.URI] === Tag.ActionScenceSwitch) {
            var sceneId = params['automationRuleId'];
            var status = params['switchStatus'];
            if (sceneId && (status === 0 || status === 1)) {
                Log.i(LOGTAG, 'operateScene sceneId: ' + sceneId + ', status: ' + status);
                var operation = status === 1 ? 'enable' : 'disable';
                self.proxy.operateScene(sceneId, operation, function (err) {
                    var code = Tag.RUN_SUCCESS;
                    var msg = "success";
                    if (err) {
                        returnErr = err;
                        code = Tag.RUN_FAIL;
                        msg = err.toString();
                    }
                    actionsLog.push(self.generateLog(code, msg, action[Tag.URI]));
                    cnt++;
                    if (!hasReturned && (cnt === self.actions.length)) {
                        hasReturned = true;
                        return callback(returnErr, actionsLog);
                    }
                });
            } else {
                cnt++;
                var mess = 'action execute　sid = ' + self.sceneId + action[Tag.URI] + ' failed for: automationRuleId =' + sceneId + ', switchStatus = ' + status;
                Log.w(LOGTAG, mess);
                returnErr = true;
                actionsLog.push(self.generateLog(Tag.RUN_FAIL, mess, action[Tag.URI]));
            }
        } else if (action[Tag.URI] === Tag.ActionNotificationIotMesssage) {
            var content = params['content'];
            if (content) {
                function notifyIotMessage() {
                    var message = {
                        sceneId: self.sceneId,
                        timestamp: Date.now(),
                        content: content
                    };
                    Log.i(LOGTAG, "notifyIotMessage ", JSON.stringify(message));
                    self.proxy.notifyIotMessage(message, function(err) {
                        var code = Tag.RUN_SUCCESS;
                        var msg = "success";
                        if(err) {
                            returnErr = err;
                            code = Tag.RUN_FAIL;
                            msg = err.toString();
                        }
                        actionsLog.push(self.generateLog(code, msg, action[Tag.URI]));
                        cnt++;
                        if(!hasReturned && (cnt === self.actions.length)) {
                            hasReturned = true;
                            return callback(returnErr, actionsLog);
                        }
                    });
                }
                Log.i(LOGTAG, "notifyIotMessage in " + delay + " seconds.");
                setTimeout(notifyIotMessage, delay * 1000);
            } else {
                cnt++;
                var mess = 'action execute　sid = ' + self.sceneId + action[Tag.URI] + ' failed for: content =' + content;
                Log.w(LOGTAG, mess);
                returnErr = true;
                actionsLog.push(self.generateLog(Tag.RUN_FAIL, mess, action[Tag.URI]));
            }
        } else {
            cnt++;
            var mess = action[Tag.URI] + ' is not support yet.';
            Log.w(LOGTAG, mess);
            returnErr = true;
            actionsLog.push(self.generateLog(Tag.RUN_FAIL, mess, action[Tag.URI]));
        }
    });
    if(cnt === this.actions.length && !hasReturned) {
        hasReturned = true;
        return callback(returnErr, actionsLog);
    }
    // Set action execution timeout.
    var timeout = maxDelay + 5;
    Log.i('Set action execution timeout (' + timeout + 's).');
    processTimeout = setTimeout(function () {
        if (hasReturned) {
            return process.exit(-1);
        }
        Log.w('Action execution timed out (' + timeout + 's), return and exit...');
        hasReturned = true;
        // Send logs, which takes only 3 seconds, and exit,
        setTimeout(function () {
            process.exit(-1);
        }, 3000);
        return callback(new Error('Action execution timed out, ' + timeout + 's.'));
    }, timeout * 1000);
};

function findTargetScene(sid, callback) {
    if(!sid) {
        return callback(new Error('"sid" can not be ' + sid));
    }
    var filePath = path.join(Tag.FOTA_IFTTT, 'configuration.json');
    fs.readFile(filePath, 'utf8', function (err, data) {
        if (err) {
            return callback(err);
        }
        var configuration;
        try {
            configuration = JSON.parse(data);
        } catch (e) {
            return callback('Configuration is in not JSON format.');
        }
        var sceneObj = configuration[sid];
        if (!sceneObj) {
            return callback(new Error('Target scene is not found: ' + sid));
        }
        callback(null, sceneObj);
    });
}

module.exports = Action;
